/*
    Kiwi
    Random number generation (for Zobrist hashing)

    Copyright (c) 1999-2004,2005 Alessandro Scotti
    http://www.ascotti.org/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "random.h"

// Random numbers extracted from the S-Box of Blowfish
static unsigned const X[55] = {
    0xd1310ba6, 0x98dfb5ac, 0x2ffd72db, 0xd01adfb7,
    0xb8e1afed, 0x6a267e96, 0xba7c9045, 0xf12c7f99,
    0x24a19947, 0xb3916cf7, 0x0801f2e2, 0x858efc16,
    0x636920d8, 0x71574e69, 0xa458fea3, 0xf4933d7e,
    0x0d95748f, 0x728eb658, 0x718bcd58, 0x82154aee,
    0x7b54a41d, 0xc25a59b5, 0x9c30d539, 0x2af26013,
    0xc5d1b023, 0x286085f0, 0xca417918, 0xb8db38ef,
    0x8e79dcb0, 0x603a180e, 0x6c9e0e8b, 0xb01e8a3e,
    0xd71577c1, 0xbd314b27, 0x78af2fda, 0x55605c60,
    0xe65525f3, 0xaa55ab94, 0x57489862, 0x63e81440,
    0x55ca396a, 0x2aab10b6, 0xb4cc5c34, 0x1141e8ce,
    0xa15486af, 0x7c72e993, 0xb3ee1411, 0x636fbc2a,
    0x2ba9c55d, 0x741831f6, 0xce5c3e16, 0x9b87931e,
    0xafd6ba33, 0x6c24cf5c
};

/*
    32-bit random number generator based on lagged Fibonacci sequence
    (see D. Knuth, "The Art Of Computer Programming", vol. 2)

    The original formulation is:

    X[n] = (X[n-55] + X[n-24]) mod m

    To implement the algorithm we make use of a precomputed table that gives 
    the values of X[0]...X[54] so that we have something to start with.
    If we set j=24 and k=55 then:
        X[n] = X[n-k] + X[n-j].
    Of course we don't want to keep the whole sequence in memory and luckily
    we don't have to, because when X[55] has been generated we no longer need
    X[0]! In fact, we can just override X[0] with X[55], then X[1] with X[56]
    and so on, thus reusing always the same locations.
*/
Random::Random()
{
    for( int i=0; i<55; i++ ) x[i] = X[i];
    j = 55 - 24;
    k = 55 - 55;
}

unsigned Random::get()
{
    int result = x[j] + x[k];

    x[k] = result;

    if( ++j >= 55 ) j = 0;
    if( ++k >= 55 ) k = 0;

    return result;
}

BitBoard Random::getBitBoard()
{
    Uint32  hi = get();
    Uint32  lo = get();

    return BitBoard( lo, hi );
}
